"""Collection of mathematical operators."""
from .addition import Add, add
from .multiply import Mul
from .matmul import Matmul
from .negative import Neg, neg
from .power import Pow

from .joint import J, Iid

from .trunkation import Trunc, Trunk, trunk

from .logarithm import Log, Log10, Logn



"""Truncated exponential distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class truncexpon(Dist):
    """Truncated exponential distribution."""

    def __init__(self, b):
        Dist.__init__(self, b=b)

    def _pdf(self, x, b):
        return numpy.exp(-x)/(1-numpy.exp(-b))

    def _cdf(self, x, b):
        return (1.0-numpy.exp(-x))/(1-numpy.exp(-b))

    def _ppf(self, q, b):
        return -numpy.log(1-q+q*numpy.exp(-b))

    def _lower(self, b):
        return 0.

    def _upper(self, b):
        return b


class TruncExponential(Add):
    """
    Truncated exponential distribution.

    Args:
        upper (float, Dist):
            Location of upper threshold
        scale (float, Dist):
            Scaling parameter in the exponential distribution
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.TruncExponential(2, 4)
        >>> distribution
        TruncExponential(scale=4, shift=0, upper=2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([0.    , 0.4142, 0.8763, 1.3988, 2.    ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.6354, 0.5729, 0.5104, 0.4479, 0.3854])
        >>> distribution.sample(4).round(4)
        array([1.1891, 0.1852, 1.873 , 0.8415])
        >>> distribution.mom(1).round(4)
        0.917
    """

    def __init__(self, upper=1, scale=1, shift=0):
        self._repr = {"upper": upper, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=truncexpon((upper-shift)*1./scale)*scale, right=shift)



"""Generalized gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class generalized_gamma(Dist):
    """Generalized gamma distribution."""

    def __init__(self, a, c):
        Dist.__init__(self, a=a, c=c)

    def _pdf(self, x, a, c):
        return abs(c)* numpy.exp((c*a-1)*numpy.log(x)-x**c- special.gammaln(a))

    def _cdf(self, x, a, c):
        val = special.gammainc(a, x**c)
        cond = c + 0*val
        return numpy.where(cond > 0, val, 1-val)

    def _ppf(self, q, a, c):
        val1 = special.gammaincinv(a, q)
        val2 = special.gammaincinv(a, 1.0-q)
        ic = 1.0/c
        cond = c+0*val1
        return numpy.where(cond > 0, val1**ic, val2**ic)

    def _mom(self, k, a, c):
        return special.gamma((c+k)*1./a)/special.gamma(c*1./a)

    def _lower(self, a, c):
        return 0.


class GeneralizedGamma(Add):
    """
    Generalized gamma distribution

    Args:
        shape1 (float, Dist):
            Shape parameter 1
        shape2 (float, Dist):
            Shape parameter 2
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.GeneralizedGamma(3, 2, 2, 2)
        >>> distribution
        GeneralizedGamma(scale=2, shape1=3, shape2=2, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([4.4779, 5.0233, 5.5244, 6.1372])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.3145, 0.4016, 0.3807, 0.2624])
        >>> distribution.sample(4).round(4)
        array([5.6691, 4.1674, 7.0214, 5.2264])
        >>> distribution.mom(1).round(4)
        3.477
    """

    def __init__(self, shape1, shape2, scale, shift):
        self._repr = {
            "shape1": shape1, "shape2": shape2, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=generalized_gamma(shape1, shape2)*scale, right=shift)



"""Exponential power distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class exponential_power(Dist):
    """Exponential power distribution."""

    def __init__(self, b=1):
        Dist.__init__(self, b=b)

    def _pdf(self, x, b):
        xbm1 = x**(b-1.0)
        xb = xbm1 * x
        return numpy.exp(1)*b*xbm1 * numpy.exp(xb - numpy.exp(xb))

    def _cdf(self, x, b):
        xb = x**b
        return -numpy.expm1(-numpy.expm1(xb))

    def _ppf(self, q, b):
        return pow(numpy.log1p(-numpy.log1p(-q)), 1.0/b)

    def _lower(self, b):
        return 0.


class ExponentialPower(Add):
    """
    Exponential power distribution.

    Also known as Generalized error distribution and Generalized normal
    distribution version 1.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.ExponentialPower(2, 2, 1)
        >>> distribution
        ExponentialPower(scale=2, shape=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([1.8976, 2.2848, 2.6129, 2.9587])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.4392, 0.5823, 0.6182, 0.5111])
        >>> distribution.sample(4).round(4)
        array([2.7003, 1.679 , 3.3551, 2.4223])
        >>> distribution.mom(1).round(4)
        2.4314
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=exponential_power(shape)*scale, right=shift)



"""Fisk or Log-logistic distribution."""
from ..baseclass import Dist
from ..operators.addition import Add


class fisk(Dist):
    """Fisk or Log-logistic distribution."""

    def __init__(self, c=1.):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        output = c*x**(c-1.)
        output /= (1+x**c)**2
        return output

    def _cdf(self, x, c):
        return 1./(1+x**-c)

    def _ppf(self, q, c):
        return (q**(-1.0)-1)**(-1.0/c)

    def _lower(self, c):
        return 0.


class Fisk(Add):
    """
    Fisk or Log-logistic distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Fisk(3, 2, 1)
        >>> distribution
        Fisk(scale=2, shape=3, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.2599, 2.7472, 3.2894, 4.1748])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.381 , 0.4121, 0.3145, 0.1512])
        >>> distribution.sample(4).round(4)
        array([3.4714, 2.013 , 6.3474, 2.9531])
        >>> distribution.mom(1).round(4)
        3.4184
    """
    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=fisk(shape)*scale, right=shift)



"""Alpha probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class alpha(Dist):
    """Standard Alpha distribution."""

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _cdf(self, x, a):
        return special.ndtr(a-1./x) / special.ndtr(a)

    def _ppf(self, q, a):
        return 1.0/(a-special.ndtri(q*special.ndtr(a)))

    def _pdf(self, x, a):
        return (1.0/(x**2)/special.ndtr(a)*
            numpy.e**(.5*(a-1.0/x)**2)/numpy.sqrt(2*numpy.pi))

    def _lower(self, a):
        return 0.


class Alpha(Add):
    """
    Alpha distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scale Parameter
        shift (float, Dist):
            Location of lower threshold

    Examples:
        >>> distribution = chaospy.Alpha(2, 0.5, 4)
        >>> distribution
        Alpha(scale=0.5, shape=2, shift=4)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([4.1676, 4.2039, 4.2465, 4.3104, 4.4521])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([11.7723,  5.4345,  3.361 ,  2.2848,  1.4892])
        >>> distribution.sample(4).round(4)
        array([4.304 , 4.1556, 4.9362, 4.2413])
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=alpha(shape)*scale, right=shift)



r"""
Archimedean copula definition.

All Copulas can be converted into Cumulative Distribution Functions (CDF) using
the following formula:

.. math::
  \begin{eqnarray}
    F(x_1) &= \frac{C(x_1, 1, 1, ...)}{C(1, 1, 1)} \\
    F(x_2\mid x_1) &= \frac{d/dx_1 C(x_1, 1, 1, ...)}{d/dx_1 C(1, 1, 1, ...)} \\
    F(x_3\mid x_1,x_2) &=
        \frac{d/dx_1 d/dx_2 C(x_1, x_2, x_3, 1, ...) }{
              d/dx_1 d/dx_2 C(x_1, x_2, 1, 1, ...)}
  \end{eqnarray}

And thereby density functions:

.. math::
  \begin{eqnarray}
    P(x_1) &= \frac{d/dx_1 C(x_1, 1, 1, ...)}{C(1, 1, 1)} \\
    P(x_2\mid x_1) &= \frac{d/dx_1 d/dx_2 C(x_1, 1, 1, ...)}{d/dx_1 C(1, 1, 1, ...)} \\
    P(x_3\mid x_1,x_2) &=
        \frac{d/dx_1 d/dx_2 d/dx_3 C(x_1, x_2, x_3, 1, ...) }{
              d/dx_1 d/dx_2 C(x_1, x_2, 1, 1, ...)}
  \end{eqnarray}

The general definition of an Archemedean is:

.. math::
    C(x_1, x_2, x_3, ...) = iphi( phi(x_1) + phi(x_2) + phi(x_3) + ... )

Applying the Copula to CDF formula on the Archemedean, we need various partial
derivatives:

.. math::
  \begin{eqnarray}
    C &= iphi(phi(x_1)+phi(x_2)+phi(x_3)+...) \\
    d/dx_1 C &= iphi'( phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) \\
    d/dx_1 d/dx_2 C &= iphi''(phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) phi'(x_2) \\
    d/dx_1 d/dx_2 d/dx_3 C &= iphi'''(phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) phi'(x_2) phi'(x_3) \\
    & \vdots \\
    d/dx_1 \cdots d/dx_n C &= iphi^{(n)}(phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) \cdots phi'(x_n)
  \end{eqnarray}

As a helper function, we introduce sigma:

.. math::
    \sigma(r, n, \theta) = \prod_{d=0}^{n-1}(-1/\theta-d)*r^{-1/theta-n}

It has the convenient property:

.. math::
    \sigma'(r, n, \theta) = \prod_{d=0}^{n}(-1/\theta-d)*r^{-1/theta-n-1} = \sigma(r, n+1, \theta)
"""
import numpy

from ..baseclass import Dist


class Archimedean(Dist):

    def __init__(self, length, theta=1.):
        self.length = length
        Dist.__init__(self, theta=float(theta))

    def __len__(self):
        return self.length

    def _lower(self, theta):
        return 0.

    def _upper(self, theta):
        return 1.

    def _cdf(self, x_loc, theta):
        out = numpy.zeros(x_loc.shape)
        loc = numpy.ones(x_loc.shape)
        for order in range(self.length):
            out[order] = self._copula(loc, theta, order)
            loc[order] = x_loc[order]
            index = out[order] != 0
            out[order, index] = self._copula(loc, theta, order)[index]/out[order, index]
            out[order, ~index] = 1
        return out

    def _pdf(self, x_loc, theta):
        out = numpy.zeros(x_loc.shape)
        loc = numpy.ones(x_loc.shape)
        for order in range(self.length):
            out[order] = self._copula(loc, theta, order)
            loc[order] = x_loc[order]
            index = out[order] != 0
            out[order, index] = self._copula(loc, theta, order+1)[index]/out[order, index]
            out[order, ~index] = 1
        return out

    def _copula(self, x_loc, theta, order=0):
        out = numpy.sum(self._phi(x_loc, theta), 0)
        out = self._inverse_phi(out, theta, order)
        if order:
            out *= numpy.where(
                out, numpy.prod(self._delta_phi(x_loc[:order], theta), 0), 0)
        else:
            out = numpy.clip(out, 0, 1)
        return out

    @staticmethod
    def _sigma(u_loc, theta, order):
        out = 1.
        for dim in range(order):
            out *= (-1/theta-dim)
        out = out*u_loc**(-1/theta-order)
        return out



"""Folded normal distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class folded_normal(Dist):
    """Folded normal distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return numpy.sqrt(2.0/numpy.pi)*numpy.cosh(c*x)*numpy.exp(-(x*x+c*c)/2.0)

    def _cdf(self, x, c):
        return special.ndtr(x-c) + special.ndtr(x+c) - 1.0

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 7.5+c


class FoldedNormal(Add):
    """
    Folded normal distribution.

    Args:
        mu (float, Dist):
            Location parameter in normal distribution
        sigma (float, Dist):
            Scaling parameter (in both normal and fold)
        loc (float, Dist):
            Location of fold

    Examples:
        >>> distribution = chaospy.FoldedNormal(3, 2, 1)
        >>> distribution
        FoldedNormal(loc=1, mu=3, sigma=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([3.3224, 4.4938, 5.5067, 6.6832])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.1417, 0.1934, 0.1932, 0.14  ])
        >>> distribution.sample(4).round(4)
        array([2.1633, 1.7827, 4.6911, 5.5156])
        >>> distribution.mom(1).round(4)
        5.034
    """

    def __init__(self, mu=0, sigma=1, loc=0):
        self._repr = {"mu": mu, "sigma": sigma, "loc": loc}
        Add.__init__(self, left= folded_normal(mu-loc)*sigma, right=loc)



"""Uniform probability distribution."""
from ..baseclass import Dist
from ..operators.addition import Add


class uniform(Dist):
    """Uniform distribution fixed on the [-1, 1] interval."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return 0.5

    def _cdf(self, x):
        return .5*x+.5

    def _ppf(self, q):
        return 2*q-1

    def _lower(self):
        return -1.

    def _upper(self):
        return 1.

    def _mom(self, k):
        return 1./(k + 1)* (k % 2 == 0)

    def _ttr(self, n):
        return 0., n*n/(4.*n*n-1)


class Uniform(Add):
    r"""
    Uniform probability distribution.

    Args:
        lower (float, Dist):
            Lower threshold of distribution. Must be smaller than ``upper``.
        upper (float, Dist):
            Upper threshold of distribution.

    Examples:
        >>> distribution = chaospy.Uniform(2, 4)
        >>> distribution
        Uniform(lower=2, upper=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([2. , 2.5, 3. , 3.5, 4. ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.5, 0.5, 0.5, 0.5, 0.5])
        >>> distribution.sample(4).round(4)
        array([3.3072, 2.23  , 3.9006, 2.9644])
        >>> distribution.mom(1).round(4)
        3.0
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[3.    , 3.    , 3.    ],
               [0.3333, 0.2667, 0.2571]])
    """

    def __init__(self, lower=0, upper=1):
        self._repr = {"lower": lower, "upper": upper}
        left = uniform()*((upper-lower)*.5)
        right = 0.5*(upper+lower)
        Add.__init__(self, left=left, right=right)



"""Fatigue-life distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class fatigue_life(Dist):
    """Fatigue-life distribution."""

    def __init__(self, c=0):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        output = (x+1)/(2*c*numpy.sqrt(2*numpy.pi*x**3))
        output *= numpy.exp(-(x-1)**2/(2.0*x*c**2))
        output[(x == 0) & numpy.isnan(output)] = 0
        return output

    def _cdf(self, x, c):
        return special.ndtr(1.0/c*(numpy.sqrt(x)-1.0/numpy.sqrt(x)))

    def _ppf(self, q, c):
        tmp = c*special.ndtri(q)
        return 0.25*(tmp + numpy.sqrt(tmp**2 + 4))**2

    def _lower(self, c):
        return 0.


class FatigueLife(Add):
    """
    Fatigue-Life or Birmbaum-Sanders distribution

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.FatigueLife(2, 2, 1)
        >>> distribution
        FatigueLife(scale=2, shape=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([ 1.4332,  2.2113,  4.3021, 10.2334])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.4223, 0.1645, 0.0603, 0.0198])
        >>> distribution.sample(4).round(4)
        array([ 5.3231,  1.2621, 26.5603,  2.8292])
        >>> distribution.mom(1).round(4)
        7.0
    """
    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=fatigue_life(shape)*scale, right=shift)



"""Log-Normal probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class log_normal(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        out = numpy.e**(-numpy.log(x+(1-x)*(x<=0))**2/(2*a*a)) / \
            ((x+(1-x)*(x<=0))*a*numpy.sqrt(2*numpy.pi))*(x>0)
        return out

    def _cdf(self, x, a):
        return special.ndtr(numpy.log(x+(1-x)*(x<=0))/a)*(x>0)

    def _ppf(self, x, a):
        return numpy.e**(a*special.ndtri(x))

    def _mom(self, k, a):
        return numpy.e**(.5*a*a*k*k)

    def _ttr(self, n, a):
        """Stieltjes-Wigert."""
        return \
    (numpy.e**(n*a*a)*(numpy.e**(a*a)+1)-1)*numpy.e**(.5*(2*n-1)*a*a), \
                (numpy.e**(n*a*a)-1)*numpy.e**((3*n-2)*a*a)

    def _lower(self, a):
        return 0.


class LogNormal(Add):
    R"""
    Log-normal distribution

    Args:
        mu (float, Dist):
            Mean in the normal distribution.  Overlaps with scale by
            mu=log(scale)
        sigma (float, Dist):
            Standard deviation of the normal distribution.
        shift (float, Dist):
            Location of the lower bound.
        scale (float, Dist):
            Scale parameter. Overlaps with mu by scale=e**mu

    Examples:
        >>> distribution = chaospy.LogNormal(0, 1)
        >>> distribution
        LogNormal(mu=0, scale=1, shift=0, sigma=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.431 , 0.7762, 1.2883, 2.3201])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.6495, 0.4977, 0.2999, 0.1207])
        >>> distribution.sample(4).round(4)
        array([1.4844, 0.3011, 5.1945, 0.9563])
        >>> distribution.mom(1).round(4)
        1.6487
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[1.50155000e+01, 1.18650900e+02, 8.97651100e+02],
               [4.67080000e+00, 3.48830600e+02, 2.09298326e+04]])
    """

    def __init__(self, mu=0, sigma=1, shift=0, scale=1):
        self._repr = {"mu": mu, "sigma": sigma, "shift": shift, "scale": scale}
        left = log_normal(sigma)*scale*numpy.e**mu
        Add.__init__(self, left=left, right=shift)


class Gilbrat(Add):
    """
    Gilbrat distribution.

    Standard log-normal distribution

    Args:
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Gilbrat(3, 2)
        >>> distribution
        Gilbrat(scale=3, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([3.293 , 4.3286, 5.865 , 8.9604])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.2165, 0.1659, 0.1   , 0.0402])
        >>> distribution.sample(4).round(4)
        array([ 6.4533,  2.9033, 17.5835,  4.869 ])
        >>> distribution.mom(1).round(4)
        6.9462
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[4.70464000e+01, 3.57952700e+02, 2.69495320e+03],
               [4.20370000e+01, 3.13947580e+03, 1.88368494e+05]])
    """

    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=log_normal(1)*scale, right=shift)



"""Power normal or Box-Cox distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class power_normal(Dist):
    """Power normal or Box-Cox distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        norm = (2*numpy.pi)**-.5*numpy.exp(-x**2/2.)
        return c*norm*special.ndtr(-x)**(c-1.)

    def _cdf(self, x, c):
        return 1.-special.ndtr(-x)**c

    def _ppf(self, q, c):
        return -special.ndtri(pow(1-q, 1./c))


class PowerNormal(Add):
    """
    Power normal or Box-Cox distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        mu (float, Dist):
            Mean of the normal distribution
        scale (float, Dist):
            Standard deviation of the normal distribution

    Examples:
        >>> distribution = chaospy.PowerNormal(2, 2, 2)
        >>> distribution
        PowerNormal(mu=2, scale=2, shape=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([-0.5008,  0.4919,  1.3233,  2.2654])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.1633, 0.2325, 0.2383, 0.1768])
        >>> distribution.sample(4).round(4)
        array([ 1.5523, -1.122 ,  3.5244,  0.8368])
        >>> distribution.mom(1).round(4)
        0.8716
    """

    def __init__(self, shape=1, mu=0, scale=1):
        self._repr = {"shape": shape, "mu": mu, "scale": scale}
        Add.__init__(self, left=power_normal(shape)*scale, right=mu)



"""Multivariate Normal Distribution."""
import numpy
from scipy import special

from .normal import normal

from ..baseclass import Dist


class MvNormal(Dist):
    """
    Multivariate Normal Distribution.

    Args:
        loc (float, Dist):
            Mean vector
        scale (float, Dist):
            Covariance matrix or variance vector if scale is a 1-d vector.

    Examples:
        >>> distribution = chaospy.MvNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> distribution
        MvNormal(loc=[1.0, 2.0], scale=[[1.0, 0.6], [0.6, 1.0]])
        >>> chaospy.Cov(distribution)
        array([[1. , 0.6],
               [0.6, 1. ]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> distribution.inv(mesh).round(4)
        array([[[0.3255, 1.    , 1.6745],
                [0.3255, 1.    , 1.6745],
                [0.3255, 1.    , 1.6745]],
        <BLANKLINE>
               [[1.0557, 1.4604, 1.8651],
                [1.5953, 2.    , 2.4047],
                [2.1349, 2.5396, 2.9443]]])
        >>> distribution.fwd(distribution.inv(mesh)).round(4)
        array([[[0.25, 0.5 , 0.75],
                [0.25, 0.5 , 0.75],
                [0.25, 0.5 , 0.75]],
        <BLANKLINE>
               [[0.25, 0.25, 0.25],
                [0.5 , 0.5 , 0.5 ],
                [0.75, 0.75, 0.75]]])
        >>> distribution.pdf(distribution.inv(mesh)).round(4)
        array([[0.0991, 0.146 , 0.1452],
               [0.1634, 0.1989, 0.1634],
               [0.1452, 0.146 , 0.0991]])
        >>> distribution.sample(4).round(4)
        array([[ 1.395 , -0.2003,  2.6476,  0.9553],
               [ 3.1476,  0.6411,  1.5946,  1.7647]])
        >>> distribution.mom((1, 2)).round(4)
        7.4
    """

    def __init__(self, loc=[0, 0], scale=[[1, .5], [.5, 1]]):
        loc = numpy.asfarray(loc)
        scale = numpy.asfarray(scale)
        assert len(loc) == len(scale)
        self._repr = {"loc": loc.tolist(), "scale": scale.tolist()}

        C = numpy.linalg.cholesky(scale)
        Ci = numpy.linalg.inv(C)
        Dist.__init__(self, C=C, Ci=Ci, loc=loc)

    def _cdf(self, x, C, Ci, loc):
        return special.ndtr(numpy.dot(Ci, (x.T-loc.T).T))

    def _ppf(self, q, C, Ci, loc):
        return (numpy.dot(C, special.ndtri(q)).T+loc.T).T

    def _pdf(self, x, C, Ci, loc):
        det = numpy.linalg.det(numpy.dot(C,C.T))
        x_ = numpy.dot(Ci.T, (x.T-loc.T).T)
        out = numpy.ones(x.shape)
        out[0] =  numpy.e**(-.5*numpy.sum(x_*x_, 0))/numpy.sqrt((2*numpy.pi)**len(Ci)*det)
        return out

    def _lower(self, C, Ci, loc):
        return -7.5*numpy.sqrt(numpy.diag(numpy.dot(C, C.T)))+loc

    def _upper(self, C, Ci, loc):
        return 7.5*numpy.sqrt(numpy.diag(numpy.dot(C, C.T)))+loc

    def _mom(self, k, C, Ci, loc):
        scale = numpy.dot(C, C.T)
        out = 0.
        for idx, kdx in enumerate(numpy.ndindex(*[_+1 for _ in k])):
            coef = numpy.prod(special.comb(k.T, kdx).T, 0)
            diff = k.T - kdx
            pos = diff >= 0
            diff = diff*pos
            pos = numpy.all(pos)
            loc_ = numpy.prod(loc**diff)

            out += pos*coef*loc_*isserlis_moment(tuple(kdx), scale)

        return float(out)

    def __len__(self):
        return len(self.prm["C"])


def isserlis_moment(k, scale):
    """
    Raw statistical moments centralized Normal distribution using Isserlis'
    theorem.

    Args:
        k (Tuple[int, ...]):
            Moment orders.
        scale (ndarray):
            Covariance matrix defining dependencies between variables.

    Returns:
        Raw statistical moment of order ``k`` given covariance ``scale``.

    Examples:
        >>> scale = 0.5*numpy.eye(3)+0.5
        >>> isserlis_moment((2, 2, 2), scale)
        3.5
        >>> isserlis_moment((0, 0, 0), scale)
        1.0
        >>> isserlis_moment((1, 0, 0), scale)
        0.0
        >>> isserlis_moment((0, 1, 1), scale)
        0.5
        >>> isserlis_moment((0, 0, 2), scale)
        1.0
    """
    if not isinstance(k, numpy.ndarray):
        k = numpy.asarray(k)
    assert len(k.shape) == 1

    # Recursive exit condition
    if (numpy.sum(k) % 2 == 1) or numpy.any(k < 0):
        return 0.

    # Choose a pivot index as first non-zero entry
    idx = numpy.nonzero(k)[0]
    if not idx.size:
        # Skip if no pivot found
        return 1.
    idx = idx[0]

    eye = numpy.eye(len(k), dtype=int)
    out = (k[idx]-1)*scale[idx, idx]*isserlis_moment(k-2*eye[idx], scale)
    for idj in range(idx+1, len(k)):
        out += k[idj]*scale[idx, idj]*isserlis_moment(
            k-eye[idx]-eye[idj], scale)

    return float(out)



import numpy

from ..baseclass import Dist
from .. import evaluation


class BinaryOperator(Dist):

    matrix = False

    def __init__(self, left, right):
        """
        Args:
            left (Dist, numpy.ndarray):
                Left hand side.
            right (Dist, numpy.ndarray):
                Right hand side.
        """
        Dist.__init__(self, left=left, right=right)

    def _precedence_order(self):
        """Precedence order of the various dimensions."""
        left = self.prm["left"]
        right = self.prm["right"]
        if isinstance(left, Dist):
            indices = left._precedence_order()
            if isinstance(right, Dist):
                assert indices == right._precedence_order()
        elif isinstance(right, Dist):
            indices = right._precedence_order()
        else:
            indices = list(range(len(self)))
        return indices


    def _cdf(self, xloc, left, right, cache):
        """Cumulative distribution function."""
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))

            if not self.matrix:
                right = (numpy.asfarray(right).T+numpy.zeros(xloc.shape).T).T
            xloc = self._pre_fwd_right(xloc, right)
            uloc = evaluation.evaluate_forward(left, xloc, cache=cache)
            uloc = self._post_fwd(uloc, right)

        elif not isinstance(right, Dist):
            uloc = self._alt_fwd(xloc, left, right)

        else:
            if not self.matrix:
                left = (numpy.asfarray(left).T+numpy.zeros(xloc.shape).T).T
            xloc = self._pre_fwd_left(xloc, left)
            uloc = evaluation.evaluate_forward(right, xloc, cache=cache)
            uloc = self._post_fwd(uloc, left)

        assert uloc.shape == xloc.shape
        return uloc



"""Truncated normal distribution."""
import numpy
from scipy import special

from ..baseclass import Dist


class TruncNormal(Dist):
    """
    Truncated normal distribution

    Args:
        lower (float, Dist):
            Location of lower threshold
        upper (float, Dist):
            Location of upper threshold
        mu (float, Dist):
            Mean of normal distribution
        sigma (float, Dist):
            Standard deviation of normal distribution

    Examples:
        >>> distribution = chaospy.TruncNormal(2, 4, 2, 2)
        >>> distribution
        TruncNormal(lower=2, mu=2, sigma=2, upper=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([2.    , 2.4311, 2.8835, 3.387 , 4.    ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([1.1687, 1.1419, 1.0601, 0.9189, 0.7089])
        >>> distribution.sample(4).round(4)
        array([3.1841, 2.1971, 3.8643, 2.8501])
    """

    def __init__(self, lower=-1, upper=1, mu=0, sigma=1):
        self._repr = {"lower": lower, "upper": upper, "mu": mu, "sigma": sigma}
        Dist.__init__(self, a=lower, b=upper, sigma=sigma, mu=mu)

    def _pdf(self, x, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        x = (x-mu)/sigma
        norm = (2*numpy.pi)**(-.5)*numpy.e**(-x**2/2.)
        return norm/(fb-fa)

    def _cdf(self, x, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        x = special.ndtr((x-mu)/sigma)
        return (x - fa) / (fb-fa)

    def _ppf(self, q, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        return special.ndtri(q*(fb-fa) + fa)*sigma + mu

    def _lower(self, a, b, mu, sigma):
        return a

    def _upper(self, a, b, mu, sigma):
        return b



"""Non-central Chi-squared distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class chi_squared(Dist):
    """Central Chi-squared distribution."""

    def __init__(self, df, nc):
        Dist.__init__(self, df=df, nc=nc)

    def _pdf(self, x, df, nc):
        output = 0.5*numpy.e**(-0.5*(x+nc))
        output *= (x/nc)**(0.25*df-0.5)
        output *= special.iv(0.5*df-1, (nc*x)**0.5)
        return output

    def _cdf(self, x, df, nc):
        return special.chndtr(x, df, nc)

    def _ppf(self, q, df, nc):
        return special.chndtrix(q, df, nc)

    def _lower(self, df, nc):
        return 0.

    def _upper(self, df, nc):
        for expon in range(20, -1, -1):
            upper = self._ppf(1-10**-expon, df, nc)
            if not numpy.isnan(upper).item():
                return upper.item()


class ChiSquared(Add):
    """
    (Non-central) Chi-squared distribution.

    Args:
        df (float, Dist):
            Degrees of freedom
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter
        nc (float, Dist):
            Non-centrality parameter

    Examples:
        >>> distribution = chaospy.ChiSquared(2, 4, 1, 1)
        >>> distribution
        ChiSquared(df=2, nc=1, scale=4, shift=1)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([ 3.369 ,  6.1849,  9.7082, 14.5166, 22.4295])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.065 , 0.0536, 0.0414, 0.0286, 0.0149])
        >>> distribution.sample(4).round(4)
        array([14.0669,  2.595 , 35.6294,  9.2851])
        >>> distribution.mom(1).round(4)
        13.0001
    """

    def __init__(self, df=1, scale=1, shift=0, nc=0):
        self._repr = {"df": df, "scale": scale, "shift": shift, "nc": nc}
        Add.__init__(self, left=chi_squared(df, nc)*scale, right=shift)



"""Laplace Probability Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add



class laplace(Dist):
    """Laplace Probability Distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return numpy.e**-numpy.abs(x)/2

    def _cdf(self, x):
        return (1+numpy.sign(x)*(1-numpy.e**-abs(x)))/2

    def _mom(self, k):
        return .5*special.factorial(k)*(1+(-1)**k)

    def _ppf(self, x):
        return numpy.where(x>.5, -numpy.log(2*(1-x)), numpy.log(2*x))

    def _lower(self):
        return -32.

    def _upper(self):
        return 32.

    def _ttr(self, k):
        from ...quadrature import quad_fejer, discretized_stieltjes
        q1, w1 = quad_fejer(500, (-32, 0))
        q2, w2 = quad_fejer(500, (0, 32))
        q = numpy.concatenate([q1,q2], 1)
        w = numpy.concatenate([w1,w2])*self.pdf(q[0])

        coeffs, _, _ = discretized_stieltjes(k, q, w)
        return coeffs[:, 0, -1]


class Laplace(Add):
    R"""
    Laplace Probability Distribution

    Args:
        mu (float, Dist):
            Mean of the distribution.
        scale (float, Dist):
            Scaling parameter. scale > 0.

    Examples:
        >>> distribution = chaospy.Laplace(2, 2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.1674, 1.5537, 2.4463, 3.8326])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.1, 0.2, 0.2, 0.1])
        >>> distribution.sample(4).round(4)
        array([ 2.734 , -0.9392,  6.6165,  1.9275])
        >>> distribution.mom(1).round(4)
        2.0
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[ 2.    ,  2.    ,  2.    ],
               [ 8.    , 39.9995, 86.4011]])
    """
    def __init__(self, mu=0, scale=1):
        self._repr = {"mu": mu, "scale": scale}
        Add.__init__(self, left=laplace()*scale, right=mu)



r"""
A cumulative distribution function of an independent multivariate random
variable can be made dependent through a copula as follows:

.. math::
    F_{Q_0,\dots,Q_{D-1}} (q_0,\dots,q_{D-1}) =
    C(F_{Q_0}(q_0), \dots, F_{Q_{D-1}}(q_{D-1}))

where :math:`C` is the copula function, and :math:`F_{Q_i}` are marginal
distribution functions.  One of the more popular classes of copulas is the
Archimedean copulas.
.. \cite{sklar_random_1996}.
They are defined as follows:

.. math::
    C(u_1,\dots,u_n) =
    \phi^{[-1]} (\phi(u_1)+\dots+\phi(u_n)),

where :math:`\phi` is a generator and :math:`\phi^{[-1]}` is its
pseudo-inverse. Support for Archimedean copulas in `chaospy` is possible
through reformulation of the Rosenblatt transformation.  In two dimension, this
reformulation is as follows:

.. math::

    F_{U_0}(u_0) = \frac{C(u_0,1)}{C(1,1)}

    F_{U_1\mid U_0}(u_1\mid u_0) =
    \frac{\tfrac{\partial}{\partial u_0}
    C(u_0,u_1)}{\tfrac{\partial}{\partial u_0} C(u_0,1)}

This definition can also be generalized in to multiple variables using the
formula provided by Nelsen 1999.
.. cite:: nelsen_introduction_1999

The definition of the Rosenblatt transform can require multiple
differentiations.  An analytical formulation is usually not feasible, so the
expressions are estimated using difference scheme similar to the one outlined
for probability density function defined in :ref:`distributions`. The accurate
might therefore be affected.

Since copulas are meant as a replacement for Rosenblatt
transformation, it is usually assumed that the distribution it is
used on is stochastically independent.
However in the definition of a copula does not actually require it, and sine
the Rosenblatt transformation allows for it, multiple copulas can be stacked
together in `chaospy`.
"""
import numpy

from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation


class Copula(Dist):

    def __init__(self, dist, trans):
        r"""
        Args:
            dist (Dist):
                Distribution to wrap the copula around.
            trans (Dist):
                The copula wrapper `[0,1]^D \into [0,1]^D`.

        """
        Dist.__init__(self, dist=dist, trans=trans)

    def _precedence_order(self):
        """Precedence order of the various dimensions."""
        dist = self.prm["dist"]
        if isinstance(dist, Dist):
            indices = dist._precedence_order()
        else:
            indices = list(range(len(self)))
        return indices

    def _cdf(self, x, dist, trans, cache):
        output = evaluation.evaluate_forward(dist, x, cache=cache)
        output = evaluation.evaluate_forward(trans, output, cache=cache)
        return output

    def _lower(self, dist, trans, cache):
        return evaluation.evaluate_lower(dist, cache=cache)

    def _upper(self, dist, trans, cache):
        return evaluation.evaluate_upper(dist, cache=cache)

    def _ppf(self, qloc, dist, trans, cache):
        qloc = evaluation.evaluate_inverse(trans, qloc, cache=cache)
        xloc = evaluation.evaluate_inverse(dist, qloc, cache=cache)
        return xloc

    def _pdf(self, x, dist, trans, cache):
        density = evaluation.evaluate_density(dist, x, cache=cache.copy())
        return evaluation.evaluate_density(
            trans, evaluation.evaluate_forward(
                dist, x, cache=cache), cache=cache)*density

    def _mom(self, x, dist, trans, cache):
        raise StochasticallyDependentError(
            "Joint distribution with dependencies not supported.")

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        args = [key + "=" + str(self._repr[key]) for key in sorted(self._repr)]
        return (self.__class__.__name__ + "(" + str(self.prm["dist"]) +
                ", " + ", ".join(args) + ")")



"""Frechet or Extreme value distribution type 2."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class frechet(Dist):
    """Frechet or Extreme value distribution type 2."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return c*pow(x,c-1)*numpy.exp(-pow(x,c))

    def _cdf(self, x, c):
        return -numpy.expm1(-pow(x,c))

    def _ppf(self, q, c):
        return pow(-numpy.log1p(-q),1.0/c)

    def _mom(self, k, c):
        return special.gamma(1-k*1./c)

    def _lower(self, c):
        return 0.


class Frechet(Add):
    """
    Frechet or Extreme value distribution type 2.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Frechet(3, 2, 1)
        >>> distribution
        Frechet(scale=2, shape=3, shift=1)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.2131, 2.5988, 2.9426, 3.3438])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.4415, 0.5751, 0.566 , 0.412 ])
        >>> distribution.sample(4).round(4)
        array([3.0393, 1.9924, 3.8849, 2.7397])
        >>> distribution.mom(1).round(4)
        3.7082
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=frechet(shape)*scale, right=shift)



"""Gumbel or Log-Weibull distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class log_weibull(Dist):
    """Gumbel or Log-Weibull distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        ex = numpy.exp(-x)
        return ex*numpy.exp(-ex)

    def _cdf(self, x):
        return numpy.exp(-numpy.exp(-x))

    def _ppf(self, q):
        return -numpy.log(-numpy.log(q))


class LogWeibull(Add):
    """
    Gumbel or Log-Weibull distribution.

    Args:
        scale (float, Dist):
            Scaling parameter
        loc (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogWeibull(2, 2)
        >>> distribution
        LogWeibull(loc=2, scale=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([1.0482, 2.1748, 3.3435, 4.9999])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.1609, 0.1833, 0.1532, 0.0893])
        >>> distribution.sample(4).round(4)
        array([3.71  , 0.4572, 7.952 , 2.631 ])
        >>> distribution.mom(1).round(4)
        3.1544
    """
    def __init__(self, scale=1, loc=0):
        self._repr = {"scale": scale, "loc": loc}
        Add.__init__(self, left=log_weibull()*scale, right=loc)



"""Nataf (normal) copula."""
import numpy
from scipy import special

from .baseclass import Copula
from ..baseclass import Dist

class nataf(Dist):
    """Nataf (normal) copula."""

    def __init__(self, R, ordering=None):
        if ordering is None:
            ordering = range(len(R))
        ordering = numpy.array(ordering)
        P = numpy.eye(len(R))[ordering]
        R = numpy.dot(P, numpy.dot(R, P.T))
        R = numpy.linalg.cholesky(R)
        R = numpy.dot(P.T, numpy.dot(R, P))
        Ci = numpy.linalg.inv(R)
        Dist.__init__(self, C=R, Ci=Ci)
        self.length = len(R)

    def __len__(self):
        return self.length

    def _cdf(self, x, C, Ci):
        out = special.ndtr(numpy.dot(Ci, special.ndtri(x)))
        return out

    def _ppf(self, q, C, Ci):
        out = special.ndtr(numpy.dot(C, special.ndtri(q)))
        return out

    def _lower(self, C, Ci):
        return 0.

    def _upper(self, C, Ci):
        return 1.


class Nataf(Copula):
    """
    Nataf (normal) copula.

    Args:
        dist (Dist):
            The Distribution to wrap.
        R (numpy.ndarray):
            Covariance matrix.

    Examples:
        >>> distribution = chaospy.Nataf(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2), R=[[1, .5], [.5, 1]])
        >>> distribution
        Nataf(Iid(Uniform(lower=-1, upper=1), 2), R=[[1, 0.5], [0.5, 1]])
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.1262,  0.3001,  0.1053]])
        >>> distribution.pdf(samples).round(4)
        array([0.292 , 0.1627, 0.2117])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.2707, -0.1737, -0.0739],
                [-0.1008,  0.    ,  0.1008],
                [ 0.0739,  0.1737,  0.2707]]])
        >>> distribution.mom([1, 1]).round(4)
        0.1609

    """

    def __init__(self, dist, R, ordering=None):
        self._repr = {"R": R}
        assert len(dist) == len(R)
        return Copula.__init__(self, dist=dist, trans=nataf(R, ordering))



import chaospy
import numpy

from ..baseclass import Dist
from .. import evaluation
from .binary import BinaryOperator


class Matmul(BinaryOperator):
    """Multiplication."""

    matrix = True

    def __init__(self, left, right):
        """
        Args:
            left (Dist, numpy.ndarray):
                Left hand side.
            right (Dist, numpy.ndarray):
                Right hand side.
        """

        if isinstance(left, Dist) and not isinstance(right, Dist):
            right = numpy.asarray(right)
            if not right.shape:
                raise ValueError("matmul: right side does not have enough dimensions")
            if len(right.shape) == 1:
                right = numpy.diag(right)

            if len(left) != right.shape[0]:
                raise ValueError("matmul: shapes incompatible (%d,) and %s" %
                                 (len(left), right.shape))

        elif not isinstance(left, Dist) and isinstance(right, Dist):
            left = numpy.asarray(left)
            if not left.shape:
                raise ValueError(
                    "matmul: left side does not have enough dimensions")
            if len(left.shape) == 1:
                left = numpy.diag(left)

            if left.shape[-1] != len(right):
                raise ValueError("matmul: shapes incompatible %s and (%d,)" %
                                 (left.shape, len(right)))

        elif not isinstance(left, Dist) and not isinstance(right, Dist):
            raise ValueError(
                "matmul: at least one argument must be a distribution")

        Dist.__init__(self, left=left, right=right)


    def _lower(self, left, right, cache):
        """Distribution lower bounds."""
        if isinstance(left, Dist):
            left_upper = evaluation.evaluate_upper(left, cache=cache)
            left_lower = evaluation.evaluate_lower(left, cache=cache)

            if isinstance(right, Dist):
                right_upper = evaluation.evaluate_upper(right, cache=cache)
                right_lower = evaluation.evaluate_lower(right, cache=cache)

                out = numpy.min(numpy.broadcast_arrays(
                    left_lower*right_lower,
                    left_lower*right_upper,
                    left_upper*right_lower,
                    left_upper*right_upper,
                ), axis=0)

            else:
                out1 = numpy.dot(left_lower.T, right).T
                out2 = numpy.dot(left_upper.T, right).T
                out = numpy.min([out1, out2], axis=0)

        elif not isinstance(right, Dist):
            out = left*right

        else:
            right_upper = evaluation.evaluate_upper(right, cache=cache)
            right_lower = evaluation.evaluate_lower(right, cache=cache)
            out = numpy.min([numpy.dot(left, right_lower),
                             numpy.dot(left, right_upper)], axis=0)

        return out

    def _upper(self, left, right, cache):
        """Distribution upper bounds."""
        if isinstance(left, Dist):
            left_lower = evaluation.evaluate_lower(left, cache=cache)
            left_upper = evaluation.evaluate_upper(left, cache=cache)

            if isinstance(right, Dist):
                right_lower = evaluation.evaluate_lower(right, cache=cache)
                right_upper = evaluation.evaluate_upper(right, cache=cache)

                out = numpy.max(numpy.broadcast_arrays(
                    (left_lower.T*right_lower.T).T,
                    (left_lower.T*right_upper.T).T,
                    (left_upper.T*right_lower.T).T,
                    (left_upper.T*right_upper.T).T,
                ), axis=0)

            else:
                out = numpy.max([numpy.dot(left_lower, right),
                                 numpy.dot(left_upper, right)], axis=0)

        elif not isinstance(right, Dist):
            out = left*right

        else:
            right_lower = evaluation.evaluate_lower(right, cache=cache)
            right_upper = evaluation.evaluate_upper(right, cache=cache)
            out = numpy.max([numpy.dot(left, right_lower),
                             numpy.dot(left, right_upper)], axis=0)

        return out

    def _pre_fwd_left(self, xloc, other):
        Ci = numpy.linalg.inv(other)
        return numpy.dot(Ci, xloc)

    def _pre_fwd_right(self, xloc, other):
        Ci = numpy.linalg.inv(other)
        return numpy.dot(xloc.T, Ci).T

    def _post_fwd(self, uloc, other):
        return uloc

    def _alt_fwd(self, xloc, left, right):
        return 0.5*(numpy.dot(left, right) == xloc)

    def _ppf(self, uloc, left, right, cache):
        """Point percentile function."""
        left = evaluation.get_inverse_cache(left, cache)
        right = evaluation.get_inverse_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))

            xloc = evaluation.evaluate_inverse(left, uloc, cache=cache)
            xloc = numpy.dot(xloc.T, right).T
            assert uloc.shape == xloc.shape

        elif not isinstance(right, Dist):
            xloc = numpy.dot(left, right)

        else:
            xloc = evaluation.evaluate_inverse(right, uloc, cache=cache)
            xloc = numpy.dot(left, xloc)

        return xloc

    def _pdf(self, xloc, left, right, cache):
        """Probability density function."""
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))

            Ci = numpy.linalg.inv(right)
            xloc = numpy.dot(xloc.T, Ci).T

            pdf = evaluation.evaluate_density(left, xloc, cache=cache)
            pdf = numpy.dot(pdf.T, Ci).T
            assert pdf.shape == xloc.shape

        elif not isinstance(right, Dist):
            pdf = numpy.inf

        else:
            Ci = numpy.linalg.inv(left)
            xloc = numpy.dot(Ci, xloc)

            pdf = evaluation.evaluate_density(right, xloc, cache=cache)
            pdf = numpy.dot(Ci, pdf)

        return numpy.abs(pdf)

    def _mom(self, key, left, right, cache):
        """Statistical moments."""
        if evaluation.get_dependencies(left, right):
            raise evaluation.DependencyError(
                "product of dependent distributions not feasible: "
                "{} and {}".format(left, right)
            )

    def _ttr(self, kloc, left, right, cache):
        """Three terms recursion coefficients."""
        raise evaluation.DependencyError("matmul causes dependencies.")

    def __str__(self):
        if self._repr is not None:
            return super(Mul, self).__str__()
        return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                ", " + str(self.prm["right"]) + ")")

    def __len__(self):
        out1 = out2 = 1
        try:
            out1 = len(self.prm["left"])
        except TypeError:
            pass
        try:
            out2 = len(self.prm["right"])
        except TypeError:
            pass
        return max(out1, out2)

    def _fwd_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left*right
        return self

    def _inv_cache(self, cache):
        left = evaluation.get_inverse_cache(self.prm["left"], cache)
        right = evaluation.get_inverse_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left*right
        return self


def mul(left, right):
    """
    Distribution multiplication.

    Args:
        left (Dist, numpy.ndarray) : left hand side.
        right (Dist, numpy.ndarray) : right hand side.
    """
    from .mv_mul import MvMul
    length = max(left, right)
    if length == 1:
        return Mul(left, right)
    return MvMul(left, right)



import logging
from functools import partial

import numpy
import chaospy

from . import evaluation


def approximate_inverse(
        distribution,
        qloc,
        parameters=None,
        cache=None,
        iterations=300,
        tol=1e-5,
        seed=None,
):
    """
    Calculate the approximation of the inverse Rosenblatt transformation.

    Uses forward Rosenblatt, probability density function (derivative) and
    boundary function to apply hybrid Newton-Raphson and binary search method.

    Args:
        distribution (Dist):
            Distribution to estimate inverse Rosenblatt.
        qloc (numpy.ndarray):
            Input values. All values must be on [0,1] and
            ``qloc.shape == (dim,size)`` where dim is the number of dimensions
            in distribution and size is the number of values to calculate
            simultaneously.
        parameters (Optional[Dict[Dist, numpy.ndarray]]):
            Parameters for the distribution.
        cache (Optional[Dict[Dist, numpy.ndarray]]):
            Memory cache for the location in the evaluation so far.
        iterations (int):
            The number of iterations allowed to be performed
        tol (float):
            Tolerance parameter determining convergence.
        seed (Optional[int]):
            Fix random seed.

    Returns:
        (numpy.ndarray):
            Approximation of inverse Rosenblatt transformation.

    Example:
        >>> distribution = chaospy.Normal(1000, 10)
        >>> qloc = numpy.array([[0.1, 0.2, 0.9]])
        >>> approximate_inverse(distribution, qloc, seed=1234).round(4)
        array([[ 987.1846,  991.5839, 1012.8152]])
        >>> distribution.inv(qloc).round(4)
        array([[ 987.1845,  991.5838, 1012.8155]])
    """
    logger = logging.getLogger(__name__)
    logger.debug("init approximate_inverse: %s", distribution)

    # lots of initial values:
    if cache is None:
        cache = {}
    xlower = distribution.lower
    xupper = distribution.upper
    xloc = 0.5*(xlower+xupper)
    xloc = (xloc.T + numpy.zeros(qloc.shape).T).T
    xlower = (xlower.T + numpy.zeros(qloc.shape).T).T
    xupper = (xupper.T + numpy.zeros(qloc.shape).T).T
    uloc = numpy.zeros(qloc.shape)
    ulower = -qloc
    uupper = 1-qloc

    for dim in distribution._precedence_order():
        indices = numpy.ones(qloc.shape[-1], dtype=bool)

        for idx in range(2*iterations):

            # evaluate function:
            uloc[dim, indices] = (evaluation.evaluate_forward(
                distribution, xloc, cache=cache.copy(),
                parameters=parameters)-qloc)[dim, indices]

            # convergence criteria:
            indices[indices] = numpy.any(numpy.abs(uloc) > tol, 0)[indices]
            if not numpy.any(indices):
                break

            # narrow down lower boundary:
            ulower[dim, indices] = numpy.where(uloc < 0, uloc, ulower)[dim, indices]
            xlower[dim, indices] = numpy.where(uloc < 0, xloc, xlower)[dim, indices]

            # narrow down upper boundary:
            uupper[dim, indices] = numpy.where(uloc > 0, uloc, uupper)[dim, indices]
            xupper[dim, indices] = numpy.where(uloc > 0, xloc, xupper)[dim, indices]

            # Newton increment every second iteration:
            xloc_ = numpy.inf
            if idx % 2 == 0:
                derivative = evaluation.evaluate_density(
                    distribution, xloc, cache=cache.copy(),
                    parameters=parameters)[dim, indices]
                derivative = numpy.where(derivative, derivative, numpy.inf)

                xloc_ = xloc[dim, indices] - uloc[dim, indices] / derivative

            # use binary search if Newton increment is outside bounds:
            weight = numpy.random.random()
            xloc[dim, indices] = numpy.where(
                (xloc_ < xupper[dim, indices]) & (xloc_ > xlower[dim, indices]),
                xloc_, (weight*xupper+(1-weight)*xlower)[dim, indices])

        else:
            logger.warning(
                "Too many iterations (dim %d) required to estimate inverse.", dim)
            logger.info("%d out of %d did not converge.",
                numpy.sum(indices), len(indices))

    logger.debug("end approximate_inverse: %s", distribution)
    return xloc


def approximate_moment(
        dist,
        K,
        retall=False,
        control_var=None,
        rule="fejer",
        order=1000,
        **kws
):
    """
    Approximation method for estimation of raw statistical moments.

    Args:
        dist : Dist
            Distribution domain with dim=len(dist)
        K : numpy.ndarray
            The exponents of the moments of interest with shape (dim,K).
        control_var : Dist
            If provided will be used as a control variable to try to reduce
            the error.
        acc (:py:data:typing.Optional[int]):
            The order of quadrature/MCI
        sparse : bool
            If True used Smolyak's sparse grid instead of normal tensor
            product grid in numerical integration.
        rule : str
            Quadrature rule.
        antithetic (:py:data:typing.Optional[numpy.ndarray]):
            List of bool. Represents the axes to mirror using antithetic
            variable during MCI.
    """
    dim = len(dist)
    shape = K.shape
    size = int(K.size/dim)
    K = K.reshape(dim, size)

    if dim > 1:
        shape = shape[1:]

    X, W = chaospy.quadrature.generate_quadrature(order, dist, rule=rule, **kws)

    grid = numpy.mgrid[:len(X[0]), :size]
    X = X.T[grid[0]].T
    K = K.T[grid[1]].T
    out = numpy.prod(X**K, 0)*W

    if control_var is not None:

        Y = control_var.ppf(dist.fwd(X))
        mu = control_var.mom(numpy.eye(len(control_var)))

        if (mu.size == 1) and (dim > 1):
            mu = mu.repeat(dim)

        for d in range(dim):
            alpha = numpy.cov(out, Y[d])[0, 1]/numpy.var(Y[d])
            out -= alpha*(Y[d]-mu)

    out = numpy.sum(out, -1)
    return out


def approximate_density(
        dist,
        xloc,
        parameters=None,
        cache=None,
        eps=1.e-7
):
    """
    Approximate the probability density function.

    Args:
        dist : Dist
            Distribution in question. May not be an advanced variable.
        xloc : numpy.ndarray
            Location coordinates. Requires that xloc.shape=(len(dist), K).
        eps : float
            Acceptable error level for the approximations
        retall : bool
            If True return Graph with the next calculation state with the
            approximation.

    Returns:
        numpy.ndarray: Local probability density function with
            ``out.shape == xloc.shape``. To calculate actual density function,
            evaluate ``numpy.prod(out, 0)``.

    Example:
        >>> distribution = chaospy.Normal(1000, 10)
        >>> xloc = numpy.array([[990, 1000, 1010]])
        >>> approximate_density(distribution, xloc).round(4)
        array([[0.0242, 0.0399, 0.0242]])
        >>> distribution.pdf(xloc).round(4)
        array([[0.0242, 0.0399, 0.0242]])
    """
    if parameters is None:
        parameters = dist.prm.copy()
    if cache is None:
        cache = {}

    xloc = numpy.asfarray(xloc)
    lo, up = numpy.min(xloc), numpy.max(xloc)
    mu = .5*(lo+up)
    eps = numpy.where(xloc < mu, eps, -eps)*xloc

    floc = evaluation.evaluate_forward(
        dist, xloc, parameters=parameters.copy(), cache=cache.copy())
    for d in range(len(dist)):
        xloc[d] += eps[d]
        tmp = evaluation.evaluate_forward(
            dist, xloc, parameters=parameters.copy(), cache=cache.copy())
        floc[d] -= tmp[d]
        xloc[d] -= eps[d]

    floc = numpy.abs(floc / eps)
    return floc



"""Log-laplace distribution."""
import numpy
from scipy import special, misc

from ..baseclass import Dist
from ..operators.addition import Add


class log_laplace(Dist):
    """Log-laplace distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        cd2 = c/2.0
        c = numpy.where(x < 1, c, -c)
        return cd2*x**(c-1)

    def _cdf(self, x, c):
        return numpy.where(x < 1, 0.5*x**c, 1-0.5*x**(-c))

    def _ppf(self, q, c):
        return numpy.where(q < 0.5, (2.*q)**(1./c), (2*(1.-q))**(-1./c))

    def _lower(self, c):
        return 0.0

class LogLaplace(Add):
    """
    Log-laplace distribution

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogLaplace(2, 2, 2)
        >>> distribution
        LogLaplace(scale=2, shape=2, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([3.2649, 3.7889, 4.2361, 5.1623])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.3162, 0.4472, 0.3578, 0.1265])
        >>> distribution.sample(4).round(4)
        array([4.4028, 2.9592, 8.3425, 3.9641])
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=log_laplace(shape)*scale, right=shift)



"""Generalized logistic type 1 distribution."""
import numpy
from scipy import special, misc

from ..baseclass import Dist
from ..operators.addition import Add

class logistic(Dist):
    """Generalized logistic type 1 distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return numpy.e**-x/(1+numpy.e**-x)**(c+1)

    def _cdf(self, x, c):
        return (1+numpy.e**-x)**-c

    def _ppf(self, q, c):
        return -numpy.log(q**(-1./c)-1)


class Logistic(Add):
    """
    Generalized logistic type 1 distribution
    Sech squared distribution

    Args:
        loc (float, Dist):
            Location parameter
        scale (float, Dist):
            Scale parameter
        skew (float, Dist):
            Shape parameter

    Examples:
        >>> distribution = chaospy.Logistic(2, 2, 2)
        >>> distribution
        Logistic(loc=2, scale=2, skew=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([1.5761, 3.0855, 4.4689, 6.2736])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.0553, 0.0735, 0.0676, 0.0422])
        >>> distribution.sample(4).round(4)
        array([4.8799, 0.6656, 9.3128, 3.6415])
        >>> distribution.mom(1).round(4)
        4.0
    """
    def __init__(self, loc=0, scale=1, skew=1):
        self._repr = {"loc": loc, "scale": scale, "skew": skew}
        Add.__init__(self, left=logistic(skew)*scale, right=loc)



"""Exponential Weibull distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class exponential_weibull(Dist):
    """Exponential Weibull distribution."""

    def __init__(self, a=1, c=1):
        Dist.__init__(self, a=a, c=c)

    def _pdf(self, x, a, c):
        exc = numpy.exp(-x**c)
        return a*c*(1-exc)**(a-1) * exc * x**(c-1)

    def _cdf(self, x, a, c):
        exm1c = -numpy.expm1(-x**c)
        return (exm1c)**a

    def _ppf(self, q, a, c):
        return (-numpy.log1p(-q**(1.0/a)))**(1.0/c)

    def _lower(self, a, c):
        return 0.


class ExponentialWeibull(Add):
    """
    Exponential Weibull distribution.

    Args:
        alpha (float, Dist):
            First shape parameter
        kappa (float, Dist):
            Second shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.ExponentialWeibull(2, 2, 2, 1)
        >>> distribution
        ExponentialWeibull(alpha=2, kappa=2, scale=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.5398, 3.0009, 3.4412, 3.9989])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.3807, 0.4651, 0.4262, 0.2832])
        >>> distribution.sample(4).round(4)
        array([3.5711, 2.2872, 4.8376, 3.1776])
        >>> distribution.mom(1).round(4)
        3.2916
    """
    def __init__(self, alpha=1, kappa=1, scale=1, shift=0):
        self._repr = {
            "alpha": alpha, "kappa": kappa, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=exponential_weibull(alpha, kappa)*scale, right=shift)



"""
Constructing custom probability distributions is done by subclassing the
distribution :class:`~chaospy.distributions.baseclass.Dist`::

    >>> class Uniform(Dist):
    ...     def __init__(self, lo=0, up=1):
    ...         '''Initializer.'''
    ...         Dist.__init__(self, lo=lo, up=up)
    ...     def _cdf(self, x_data, lo, up):
    ...         '''Cumulative distribution function.'''
    ...         return (x_data-lo)/(up-lo)
    ...     def _lower(self, lo, up):
    ...         '''Lower bound.'''
    ...         return lo
    ...     def _upper(self, lo, up):
    ...         '''Upper bound.'''
    ...         return up
    ...     def _pdf(self, x_data, lo, up):
    ...         '''Probability density function.'''
    ...         return 1./(up-lo)
    ...     def _ppf(self, q_data, lo, up):
    ...         '''Point percentile function.'''
    ...         return q_data*(up-lo) + lo

Usage is then straight forward::

    >>> dist = Uniform(-3, 3)
    >>> dist.fwd([-3, 0, 3])  # Forward Rosenblatt transformation
    array([0. , 0.5, 1. ])

Here the method ``_cdf`` is an absolute requirement. In addition, either
``_ppf``, or the couple ``_lower`` and ``_upper`` should be provided. The
others are not required, but may increase speed and or accuracy of
calculations. In addition to the once listed, it is also
possible to define the following methods:

``_mom``
    Method for creating raw statistical moments, used by the ``mom`` method.
``_ttr``
    Method for creating coefficients from three terms recursion method, used to
    perform "analytical" Stiltjes' method.
"""
import types
import numpy

from . import evaluation, approximation


class StochasticallyDependentError(Exception):
    """Error related to stochastically dependent variables."""


class Dist(object):
    """Baseclass for all probability distributions."""

    __array_priority__ = 9000
    """Numpy override variable."""
    _repr = None

    interpret_as_integer = False
    """
    Flag indicating that return value from the methods sample, and inv
    should be interpreted as integers instead of floating point.
    """

    def _precedence_order(self):
        """Precedence order of the various dimensions."""
        return list(range(len(self)))

    def _lower(self, **prm):
        """Backend lower bound."""
        return self._ppf(numpy.array([1e-10]*len(self)), **prm)

    @property
    def lower(self):
        """Lower bound for the distribution."""
        return evaluation.evaluate_lower(self)

    def _upper(self, **prm):
        """Backend upper bound."""
        return self._ppf(numpy.array([1-1e-10]*len(self)), **prm)

    @property
    def upper(self):
        """Upper bound for the distribution."""
        return evaluation.evaluate_upper(self)

    def __init__(self, **prm):
        """
        Args:
            prm (numpy.ndarray):
                Other optional parameters. Will be assumed when calling any
                sub-functions.
        """
        self.prm = prm

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        assert "_bnd" not in cls.__dict__, (
            "Dist._bnd is deprecated. Use Dist._lower and Dist._upper instead.")


    def fwd(self, x_data):
        """
        Forward Rosenblatt transformation.

        Args:
            x_data (numpy.ndarray):
                Location for the distribution function. ``x_data.shape`` must
                be compatible with distribution shape.

        Returns:
            (numpy.ndarray):
                Evaluated distribution function values, where
                ``out.shape==x_data.shape``.
        """
        x_data = numpy.asfarray(x_data)
        shape = x_data.shape
        x_data = x_data.reshape(len(self), -1)

        q_data = numpy.zeros(x_data.shape)
        indices = (x_data.T > self.upper.T).T
        q_data[indices] = 1
        indices = ~indices & (x_data.T >= self.lower).T

        q_data[indices] = numpy.clip(evaluation.evaluate_forward(
            self, x_data), a_min=0, a_max=1)[indices]

        q_data = q_data.reshape(shape)
        return q_data

    def cdf(self, x_data):
        """
        Cumulative distribution function.

        Note that chaospy only supports cumulative distribution functions for
        stochastically independent distributions.

        Args:
            x_data (numpy.ndarray):
                Location for the distribution function. Assumes that
                ``len(x_data) == len(distribution)``.

        Returns:
            (numpy.ndarray):
                Evaluated distribution function values, where output has shape
                ``x_data.shape`` in one dimension and ``x_data.shape[1:]`` in
                higher dimensions.
        """
        if len(self) > 1 and evaluation.get_dependencies(*self):
            raise StochasticallyDependentError(
                "Cumulative distribution does not support dependencies.")
        q_data = self.fwd(x_data)
        if len(self) > 1:
            q_data = numpy.prod(q_data, 0)
        return q_data

    def inv(self, q_data, max_iterations=100, tollerance=1e-5):
        """
        Inverse Rosenblatt transformation.

        If possible the transformation is done analytically. If not possible,
        transformation is approximated using an algorithm that alternates
        between Newton-Raphson and binary search.

        Args:
            q_data (numpy.ndarray):
                Probabilities to be inverse. If any values are outside ``[0,
                1]``, error will be raised. ``q_data.shape`` must be compatible
                with distribution shape.
            max_iterations (int):
                If approximation is used, this sets the maximum number of
                allowed iterations in the Newton-Raphson algorithm.
            tollerance (float):
                If approximation is used, this set the error tolerance level
                required to define a sample as converged.

        Returns:
            (numpy.ndarray):
                Inverted probability values where
                ``out.shape == q_data.shape``.
        """
        q_data = numpy.asfarray(q_data)
        assert numpy.all((q_data >= 0) & (q_data <= 1)), "sanitize your inputs!"
        shape = q_data.shape
        q_data = q_data.reshape(len(self), -1)
        x_data = evaluation.evaluate_inverse(self, q_data)
        x_data = numpy.clip(x_data.T, self.lower, self.upper).T
        x_data = x_data.reshape(shape)
        return x_data

    def pdf(self, x_data, step=1e-7):
        """
        Probability density function.

        If possible the density will be calculated analytically. If not
        possible, it will be approximated by approximating the one-dimensional
        derivative of the forward Rosenblatt transformation and multiplying the
        component parts. Note that even if the distribution is multivariate,
        each component of the Rosenblatt is one-dimensional.

        Args:
            x_data (numpy.ndarray):
                Location for the density function. ``x_data.shape`` must be
                compatible with distribution shape.
            step (float, numpy.ndarray):
                If approximation is used, the step length given in the
                approximation of the derivative. If array provided, elements
                are used along each axis.

        Returns:
            (numpy.ndarray):
                Evaluated density function values. Shapes are related through
                the identity ``x_data.shape == dist.shape+out.shape``.
        """
        x_data = numpy.asfarray(x_data)
        shape = x_data.shape
        x_data = x_data.reshape(len(self), -1)

        f_data = numpy.zeros(x_data.shape)
        indices = (x_data.T <= self.upper).T & (x_data.T >= self.lower).T
        f_data[indices] = evaluation.evaluate_density(self, x_data)[indices]
        f_data = f_data.reshape(shape)
        if len(self) > 1:
            f_data = numpy.prod(f_data, 0)
        return f_data

    def sample(self, size=(), rule="random", antithetic=None):
        """
        Create pseudo-random generated samples.

        By default, the samples are created using standard (pseudo-)random
        samples. However, if needed, the samples can also be created by either
        low-discrepancy sequences, and/or variance reduction techniques.

        Changing the sampling scheme, use the following ``rule`` flag:

        +----------------------+-----------------------------------------------+
        | key                  | Description                                   |
        +======================+===============================================+
        | ``chebyshev``        | Roots of first order Chebyshev polynomials.   |
        +----------------------+-----------------------------------------------+
        | ``nested_chebyshev`` | Chebyshev nodes adjusted to ensure nested.    |
        +----------------------+-----------------------------------------------+
        | ``korobov``          | Korobov lattice.                              |
        +----------------------+-----------------------------------------------+
        | ``random``           | Classical (Pseudo-)Random samples.            |
        +----------------------+-----------------------------------------------+
        | ``grid``             | Regular spaced grid.                          |
        +----------------------+-----------------------------------------------+
        | ``nested_grid``      | Nested regular spaced grid.                   |
        +----------------------+-----------------------------------------------+
        | ``latin_hypercube``  | Latin hypercube samples.                      |
        +----------------------+-----------------------------------------------+
        | ``sobol``            | Sobol low-discrepancy sequence.               |
        +----------------------+-----------------------------------------------+
        | ``halton``           | Halton low-discrepancy sequence.              |
        +----------------------+-----------------------------------------------+
        | ``hammersley``       | Hammersley low-discrepancy sequence.          |
        +----------------------+-----------------------------------------------+

        All samples are created on the ``[0, 1]``-hypercube, which then is
        mapped into the domain of the distribution using the inverse Rosenblatt
        transformation.

        Args:
            size (numpy.ndarray):
                The size of the samples to generate.
            rule (str):
                Indicator defining the sampling scheme.
            antithetic (bool, numpy.ndarray):
                If provided, will be used to setup antithetic variables. If
                array, defines the axes to mirror.

        Returns:
            (numpy.ndarray):
                Random samples with shape ``(len(self),)+self.shape``.
        """
        size_ = numpy.prod(size, dtype=int)
        dim = len(self)
        if dim > 1:
            if isinstance(size, (tuple, list, numpy.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        from . import sampler
        out = sampler.generator.generate_samples(
            order=size_, domain=self, rule=rule, antithetic=antithetic)
        try:
            out = out.reshape(shape)
        except:
            if len(self) == 1:
                out = out.flatten()
            else:
                out = out.reshape(dim, int(out.size/dim))

        if self.interpret_as_integer:
            out = out.astype(int)
        return out

    def mom(self, K, **kws):
        """
        Raw statistical moments.

        Creates non-centralized raw moments from the random variable. If
        analytical options can not be utilized, Monte Carlo integration
        will be used.

        Args:
            K (numpy.ndarray):
                Index of the raw moments. k.shape must be compatible with
                distribution shape.  Sampling scheme when performing Monte
                Carlo
            rule (str):
                rule for estimating the moment if the analytical method fails.
            antithetic (numpy.ndarray):
                List of bool. Represents the axes to mirror using antithetic
                variable during MCI.

        Returns:
            (numpy.ndarray):
                Shapes are related through the identity
                ``k.shape == dist.shape+k.shape``.
        """
        K = numpy.asarray(K, dtype=int)
        shape = K.shape
        dim = len(self)

        if dim > 1:
            shape = shape[1:]

        size = int(K.size/dim)
        K = K.reshape(dim, size)
        out = [evaluation.evaluate_moment(self, kdata, {}) for kdata in K.T]
        out = numpy.array(out)
        return out.reshape(shape)

    def _mom(self, *args, **kws):
        """Default moment generator, throws error."""
        raise StochasticallyDependentError("component lack support")

    def ttr(self, kloc, acc=10**3, verbose=1):
        """
        Three terms relation's coefficient generator

        Args:
            k (numpy.ndarray, int):
                The order of the coefficients.
            acc (int):
                Accuracy of discretized Stieltjes if analytical methods are
                unavailable.

        Returns:
            (Recurrence coefficients):
                Where out[0] is the first (A) and out[1] is the second
                coefficient With ``out.shape==(2,)+k.shape``.
        """
        kloc = numpy.asarray(kloc, dtype=int)
        shape = kloc.shape
        kloc = kloc.reshape(len(self), -1)
        out = [evaluation.evaluate_recurrence_coefficients(self, k) for k in kloc.T]
        alpha, beta = numpy.asfarray(list(zip(*out)))
        out = numpy.array([alpha.T, beta.T])
        return out.reshape((2,)+shape)


    def _ttr(self, kloc, cache, **kws):
        """Default TTR generator, throws error."""
        raise NotImplementedError()

    def __str__(self):
        """X.__str__() <==> str(X)"""
        if self._repr is not None:
            kwargs = self._repr
        else:
            kwargs = self.prm
        args = [str(arg) for arg in kwargs.pop("_", [])]
        args += [key + "=" + str(kwargs[key]) for key in sorted(kwargs)]
        return self.__class__.__name__ + "(" + ", ".join(args) + ")"

    def __repr__(self):
        return str(self)

    def __len__(self):
        """X.__len__() <==> len(X)"""
        return 1

    def __add__(self, X):
        """Y.__add__(X) <==> X+Y"""
        from . import operators
        return operators.Add(self, X)

    def __radd__(self, X):
        """Y.__radd__(X) <==> Y+X"""
        from . import operators
        return operators.Add(self, X)

    def __sub__(self, X):
        """Y.__sub__(X) <==> X-Y"""
        from . import operators
        return operators.Add(self, -X)

    def __rsub__(self, X):
        """Y.__rsub__(X) <==> Y-X"""
        from . import operators
        return operators.Add(X, -self)

    def __neg__(self):
        """X.__neg__() <==> -X"""
        from . import operators
        return operators.Neg(self)

    def __matmul__(self, X):
        """Y.__matmul__(X) <==> X@Y"""
        from . import operators
        return operators.Matmul(self, X)

    def __rmatmul__(self, X):
        """Y.__rmatmul__(X) <==> Y@X"""
        from . import operators
        return operators.Matmul(X, self)

    def __mul__(self, X):
        """Y.__mul__(X) <==> X*Y"""
        from . import operators
        return operators.Mul(self, X)

    def __rmul__(self, X):
        """Y.__rmul__(X) <==> Y*X"""
        from . import operators
        return operators.Mul(X, self)

    def __div__(self, X):
        """Y.__div__(X) <==> Y/X"""
        from . import operators
        return operators.Mul(self, X**-1)

    def __rdiv__(self, X):
        """Y.__rdiv__(X) <==> X/Y"""
        from . import operators
        return operators.Mul(X, self**-1)

    def __floordiv__(self, X):
        """Y.__floordiv__(X) <==> Y/X"""
        from . import operators
        return operators.Mul(self, X**-1)

    def __rfloordiv__(self, X):
        """Y.__rfloordiv__(X) <==> X/Y"""
        from . import operators
        return operators.Mul(X, self**-1)

    def __truediv__(self, X):
        """Y.__truediv__(X) <==> Y/X"""
        from . import operators
        return operators.Mul(self, X**-1)

    def __rtruediv__(self, X):
        """Y.__rtruediv__(X) <==> X/Y"""
        from . import operators
        return operators.Mul(X, self**-1)

    def __pow__(self, X):
        """Y.__pow__(X) <==> Y**X"""
        from . import operators
        return operators.Pow(self, X)

    def __rpow__(self, X):
        """Y.__rpow__(X) <==> X**Y"""
        from . import operators
        return operators.Pow(X, self)

    def __le__(self, X):
        """Y.__le__(X) <==> Y<=X"""
        from . import operators
        return operators.Trunc(self, X)

    def __lt__(self, X):
        """Y.__lt__(X) <==> Y<X"""
        from . import operators
        return operators.Trunc(self, X)

    def __ge__(self, X):
        """Y.__ge__(X) <==> Y>=X"""
        from . import operators
        return operators.Trunc(X, self)

    def __gt__(self, X):
        """Y.__gt__(X) <==> Y>X"""
        from . import operators
        return operators.Trunc(X, self)



"""Double gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class double_gamma(Dist):
    """Double gamma distribution."""

    def __init__(self, a):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        ax = abs(x)
        return 1.0/(2*special.gamma(a))*ax**(a-1.0) * numpy.exp(-ax)

    def _cdf(self, x, a):
        fac = 0.5*special.gammainc(a,abs(x))
        return numpy.where(x>0,0.5+fac,0.5-fac)

    def _ppf(self, q, a):
        fac = special.gammainccinv(a,1-abs(2*q-1))
        return numpy.where(q>0.5, fac, -fac)


class DoubleGamma(Add):
    """
    Double gamma distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.DoubleGamma(2, 4, 2)
        >>> distribution
        DoubleGamma(scale=4, shape=2, shift=2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([-100.4566,   -4.7134,    2.    ,    8.7134,  104.4566])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.    , 0.0392, 0.    , 0.0392, 0.    ])
        >>> distribution.sample(4).round(4)
        array([ 6.4679, -9.2251, 17.5874,  0.8239])
        >>> distribution.mom(1).round(4)
        2.0
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=double_gamma(shape)*scale, right=shift)



"""T-Copula."""
import numpy
from scipy import special

from .baseclass import Copula
from ..baseclass import Dist


class t_copula(Dist):
    """T-Copula."""

    def __init__(self, df, R):
        C = numpy.linalg.cholesky(R)
        Ci = numpy.linalg.inv(C)
        self.length = len(R)
        Dist.__init__(self, df=df, C=C, Ci=Ci)

    def __len__(self):
        return self.length

    def _cdf(self, x, df, C, Ci):
        out = special.stdtr(df, numpy.dot(Ci, special.stdtrit(df, x)))
        return out

    def _ppf(self, q, df, C, Ci):
        out = special.stdtr(df, numpy.dot(C, special.stdtrit(df, q)))
        return out

    def _lower(self, df, C, Ci):
        return 0.

    def _upper(self, df, C, Ci):
        return 1.


class TCopula(Copula):
    """
    T-Copula.

    Args:
        dist (Dist):
            The Distribution to wrap in a copula.
        R (numpy.ndarray):
            Covariance matrix defining dependencies..
        df (float):
            The degree of freedom in the underlying student-t distribution.

    Examples:
        >>> distribution = chaospy.TCopula(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2),
        ...     df=5, R=[[1, .5], [.5, 1]])
        >>> distribution
        TCopula(Iid(Uniform(lower=-1, upper=1), 2), R=[[1, 0.5], [0.5, 1]], df=5)
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.1274,  0.3147,  0.1928]])
        >>> distribution.pdf(samples).round(4)
        array([0.2932, 0.1367, 0.1969])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.2699, -0.1738, -0.0741],
                [-0.1011,  0.    ,  0.1011],
                [ 0.0741,  0.1738,  0.2699]]])
        >>> distribution.mom([1, 1]).round(4)
        0.1665

    """

    def __init__(self, dist, df, R):
        self._repr = {"df": df, "R": R}
        Copula.__init__(self, dist=dist, trans=t_copula(df, R))



"""
Multiplication of distributions.

Example usage
-------------

Distribution * a constant::

    >>> distribution = chaospy.Uniform(0, 1) * 4
    >>> print(distribution)
    Mul(Uniform(lower=0, upper=1), [4.])
    >>> print(numpy.around(distribution.sample(5), 4))
    [2.6144 0.46   3.8011 1.9288 3.4899]
    >>> print(numpy.around(distribution.fwd([1, 2, 3]), 4))
    [0.25 0.5  0.75]
    >>> print(numpy.around(distribution.inv(distribution.fwd([1, 2, 3])), 4))
    [1. 2. 3.]
    >>> print(numpy.around(distribution.pdf([1, 2, 3]), 4))
    [0.25 0.25 0.25]
    >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
    [ 2.      5.3333 16.    ]
    >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
    [[2.     2.     2.    ]
     [1.3333 1.0667 1.0286]]

Construct joint multiplication distribution::

    >>> lhs = chaospy.Uniform(-1, 0)
    >>> rhs = chaospy.Uniform(-3, -2)
    >>> multiplication = lhs * rhs
    >>> print(multiplication)
    Mul(Uniform(lower=-1, upper=0), Uniform(lower=-3, upper=-2))
    >>> joint1 = chaospy.J(lhs, multiplication)
    >>> print(joint1.lower)
    [-1.  0.]
    >>> print(joint1.upper)
    [0. 3.]
    >>> joint2 = chaospy.J(rhs, multiplication)
    >>> print(joint2.lower)
    [-3.  0.]
    >>> print(joint2.upper)
    [-2.  3.]
    >>> joint3 = chaospy.J(multiplication, lhs)
    >>> print(joint3.lower)
    [ 0. -1.]
    >>> print(joint3.upper)
    [3. 0.]
    >>> joint4 = chaospy.J(multiplication, rhs)
    >>> print(joint4.lower)
    [ 0. -3.]
    >>> print(joint4.upper)
    [ 3. -2.]

Generate random samples::

    >>> print(numpy.around(joint1.sample(4), 4))
    [[-0.7877 -0.9593 -0.6028 -0.7669]
     [ 2.2383  2.1172  1.6532  1.8345]]
    >>> print(numpy.around(joint2.sample(4), 4))
    [[-2.8177 -2.2565 -2.9304 -2.1147]
     [ 2.6843  2.1011  1.2174  0.0613]]

Forward transformations::

    >>> lcorr = numpy.array([-0.9, -0.5, -0.1])
    >>> rcorr = numpy.array([-2.99, -2.5, -2.01])
    >>> print(numpy.around(joint1.fwd([lcorr, lcorr*rcorr]), 4))
    [[0.1  0.5  0.9 ]
     [0.99 0.5  0.01]]
    >>> print(numpy.around(joint2.fwd([rcorr, lcorr*rcorr]), 4))
    [[0.01 0.5  0.99]
     [0.9  0.5  0.1 ]]

Inverse transformations::

    >>> print(numpy.around(joint1.inv(joint1.fwd([lcorr, lcorr*rcorr])), 4))
    [[-0.9   -0.5   -0.1  ]
     [ 2.691  1.25   0.201]]
    >>> print(numpy.around(joint2.inv(joint2.fwd([rcorr, lcorr*rcorr])), 4))
    [[-2.99  -2.5   -2.01 ]
     [ 2.691  1.25   0.201]]
"""
import chaospy
import numpy

from ..baseclass import Dist
from .. import evaluation
from .binary import BinaryOperator


class Mul(BinaryOperator):
    """Multiplication."""

    def __init__(self, left, right):
        """
        Args:
            left (Dist, numpy.ndarray):
                Left hand side.
            right (Dist, numpy.ndarray):
                Right hand side.
        """
        self.matrix = False
        if isinstance(left, Dist):
            if not isinstance(right, Dist):
                right = right*numpy.ones(len(left))
                if len(right.shape) > 1:
                    raise ValueError("Too many dimensions")
                if len(left) == 1 and len(right) > 1:
                    raise ValueError("Univariate distribution x vector")

        else:
            if isinstance(right, Dist):
                left = left*numpy.ones(len(right))
                if len(left.shape) > 1:
                    raise ValueError("Too many dimensions")
                if len(right) == 1 and len(left) > 1:
                    raise ValueError("Univariate distribution x vector")
            else:
                raise ValueError(
                    "Either left or right side must be distributions")

        Dist.__init__(self, left=left, right=right)

    def _lower(self, left, right, cache):
        """
        Distribution lower bounds.

        Example:
            >>> chaospy.Mul(chaospy.Uniform(-1, 2), -2).lower
            array([-4.])
            >>> chaospy.Mul(chaospy.Uniform(-1, 1), chaospy.Uniform(1, 2)).lower
            array([-2.])
        """
        if isinstance(left, Dist):
            left_upper = evaluation.evaluate_upper(left, cache=cache)
            left_lower = evaluation.evaluate_lower(left, cache=cache)

            if isinstance(right, Dist):
                right_upper = evaluation.evaluate_upper(right, cache=cache)
                right_lower = evaluation.evaluate_lower(right, cache=cache)

                out = numpy.min(numpy.broadcast_arrays(
                    left_lower*right_lower,
                    left_lower*right_upper,
                    left_upper*right_lower,
                    left_upper*right_upper,
                ), axis=0)

            else:
                out = numpy.min([left_lower*right, left_upper*right], axis=0)

        elif not isinstance(right, Dist):
            out = left*right

        else:
            right_upper = evaluation.evaluate_upper(right, cache=cache)
            right_lower = evaluation.evaluate_lower(right, cache=cache)
            out = numpy.min([left*right_lower, left*right_upper], axis=0)

        return out

    def _upper(self, left, right, cache):
        """
        Distribution upper bounds.

        Example:
            >>> chaospy.Mul(chaospy.Uniform(-1, 2), -2).upper
            array([2.])
            >>> chaospy.Mul(chaospy.Uniform(-1, 1), chaospy.Uniform(1, 2)).upper
            array([2.])
        """
        if isinstance(left, Dist):
            left_lower = evaluation.evaluate_lower(left, cache=cache)
            left_upper = evaluation.evaluate_upper(left, cache=cache)

            if isinstance(right, Dist):
                right_lower = evaluation.evaluate_lower(right, cache=cache)
                right_upper = evaluation.evaluate_upper(right, cache=cache)

                out = numpy.max(numpy.broadcast_arrays(
                    (left_lower.T*right_lower.T).T,
                    (left_lower.T*right_upper.T).T,
                    (left_upper.T*right_lower.T).T,
                    (left_upper.T*right_upper.T).T,
                ), axis=0)

            else:
                out = numpy.max([left_lower*right, left_upper*right], axis=0)

        elif not isinstance(right, Dist):
            out = left*right

        else:
            right_lower = evaluation.evaluate_lower(right, cache=cache)
            right_upper = evaluation.evaluate_upper(right, cache=cache)
            out = numpy.max([left*right_lower, left*right_upper], axis=0)

        return out

    def _pre_fwd_left(self, xloc, other):
        valids = other != 0
        out = xloc.copy()
        out.T[valids.T] = xloc.T[valids.T]/other.T[valids.T]
        return out

    def _pre_fwd_right(self, xloc, other):
        valids = other != 0
        out = xloc.copy()
        out.T[valids.T] = xloc.T[valids.T]/other.T[valids.T]
        return out

    def _post_fwd(self, uloc, other):
        return numpy.where(other.T >= 0, uloc.T, 1-uloc.T).T

    def _alt_fwd(self, xloc, left, right):
        return 0.5*(left*right == xloc)

    def _ppf(self, uloc, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Uniform().inv([0.1, 0.2, 0.9]))
            [0.1 0.2 0.9]
            >>> print(Mul(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9]))
            [0.2 0.4 1.8]
            >>> print(Mul(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9]))
            [0.2 0.4 1.8]
            >>> dist = chaospy.Mul([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> print(dist.inv([[0.5, 0.6, 0.7], [0.5, 0.6, 0.7]]))
            [[1.  1.2 1.4]
             [0.5 0.6 0.7]]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.inv([[0.5, 0.6, 0.7], [0.5, 0.6, 0.7]]))
            [[0.5 0.6 0.7]
             [1.  1.2 1.4]]
        """
        left = evaluation.get_inverse_cache(left, cache)
        right = evaluation.get_inverse_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))

            uloc = numpy.where(numpy.asfarray(right).T > 0, uloc.T, 1-uloc.T).T
            xloc = evaluation.evaluate_inverse(left, uloc, cache=cache)
            xloc = (xloc.T*right.T).T
            assert uloc.shape == xloc.shape

        elif not isinstance(right, Dist):
            xloc = left*right

        else:
            uloc = numpy.where(numpy.asfarray(left).T > 0, uloc.T, 1-uloc.T).T
            xloc = evaluation.evaluate_inverse(right, uloc, cache=cache)
            xloc = (xloc.T*left.T).T

        return xloc

    def _pdf(self, xloc, left, right, cache):
        """
        Probability density function.

        Example:
            >>> print(chaospy.Uniform().pdf([-0.5, 0.5, 1.5, 2.5]))
            [0. 1. 0. 0.]
            >>> print(Mul(chaospy.Uniform(), 2).pdf([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 0.5 0. ]
            >>> print(Mul(2, chaospy.Uniform()).pdf([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 0.5 0. ]
            >>> dist = chaospy.Mul([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> print(dist.pdf([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [0.5 0.5 0. ]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.pdf([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [0.5 0.5 0. ]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))

            right = (numpy.asfarray(right).T+numpy.zeros(xloc.shape).T).T
            valids = right != 0
            xloc.T[valids.T] = xloc.T[valids.T]/right.T[valids.T]
            xloc.T[~valids.T] = numpy.inf

            pdf = evaluation.evaluate_density(left, xloc, cache=cache)
            pdf.T[valids.T] /= right.T[valids.T]
            assert pdf.shape == xloc.shape

        elif not isinstance(right, Dist):
            pdf = numpy.inf

        else:
            left = (numpy.asfarray(left).T+numpy.zeros(xloc.shape).T).T
            valids = left != 0
            xloc.T[valids.T] = xloc.T[valids.T]/left.T[valids.T]

            pdf = evaluation.evaluate_density(right, xloc, cache=cache)
            pdf.T[valids.T] /= left.T[valids.T]

        return numpy.abs(pdf)

    def _mom(self, key, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> print(numpy.around(chaospy.Uniform().mom([0, 1, 2, 3]), 4))
            [1.     0.5    0.3333 0.25  ]
            >>> print(numpy.around(Mul(chaospy.Uniform(), 2).mom([0, 1, 2, 3]), 4))
            [1.     1.     1.3333 2.    ]
            >>> print(numpy.around(Mul(2, chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [1.     1.     1.3333 2.    ]
            >>> print(numpy.around(Mul(chaospy.Uniform(), chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [1.     0.25   0.1111 0.0625]
        """
        if evaluation.get_dependencies(left, right):
            raise evaluation.DependencyError(
                "product of dependent distributions not feasible: "
                "{} and {}".format(left, right)
            )

        if isinstance(left, Dist):
            left = evaluation.evaluate_moment(left, key, cache=cache)
        else:
            left = (numpy.array(left).T**key).T
        if isinstance(right, Dist):
            right = evaluation.evaluate_moment(right, key, cache=cache)
        else:
            right = (numpy.array(right).T**key).T
        return numpy.prod(left)*numpy.prod(right)

    def _ttr(self, kloc, left, right, cache):
        """Three terms recursion coefficients."""
        if isinstance(left, Dist) and isinstance(right, Dist):
            raise evaluation.DependencyError(
                "product of distributions not feasible: "
                "{} and {}".format(left, right)
            )
        elif isinstance(right, Dist):
            left, right = right, left

        coeff0, coeff1 = evaluation.evaluate_recurrence_coefficients(
            left, kloc, cache=cache)
        right = numpy.asarray(right)
        return coeff0*right, coeff1*right*right

    def __str__(self):
        if self._repr is not None:
            return super(Mul, self).__str__()
        return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                ", " + str(self.prm["right"]) + ")")

    def __len__(self):
        out1 = out2 = 1
        try:
            out1 = len(self.prm["left"])
        except TypeError:
            pass
        try:
            out2 = len(self.prm["right"])
        except TypeError:
            pass
        return max(out1, out2)

    def _fwd_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left*right
        return self

    def _inv_cache(self, cache):
        left = evaluation.get_inverse_cache(self.prm["left"], cache)
        right = evaluation.get_inverse_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left*right
        return self



"""Kumaraswswamy's double bounded distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class kumaraswamy(Dist):
    """Kumaraswswamy's double bounded distribution."""

    def __init__(self, a=1, b=1):
        assert numpy.all(a > 0) and numpy.all(b > 0)
        Dist.__init__(self, a=a, b=b)

    def _pdf(self, x, a, b):
        return a*b*x**(a-1)*(1-x**a)**(b-1)

    def _cdf(self, x, a, b):
        return 1-(1-x**a)**b

    def _ppf(self, q, a, b):
        return (1-(1-q)**(1./b))**(1./a)

    def _mom(self, k, a, b):
        return b*special.gamma(1+k*1./a)*special.gamma(b)/\
                special.gamma(1+b+k*1./a)

    def _lower(self, a, b):
        return 0.

    def _upper(self, a, b):
        return 1.


class Kumaraswamy(Add):
    """
    Kumaraswamy's double bounded distribution

    Args:
        alpha (float, Dist):
            First shape parameter, alpha > 0
        beta (float, Dist):
            Second shape parameter, b > 0
        lower (float, Dist):
            Lower threshold
        upper (float, Dist):
            Upper threshold

    Examples:
        >>> distribution = chaospy.Kumaraswamy(2, 2, 2, 3)
        >>> distribution
        Kumaraswamy(alpha=2, beta=2, lower=2, upper=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.3249, 2.4748, 2.6063, 2.7435])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([1.1625, 1.471 , 1.5337, 1.33  ])
        >>> distribution.sample(4).round(4)
        array([2.6414, 2.2434, 2.8815, 2.5295])
        >>> distribution.mom(1).round(4)
        2.5333
    """

    def __init__(self, alpha, beta, lower=0, upper=1):
        self._repr = {
            "alpha": alpha, "beta": beta, "lower": lower, "upper": upper}
        Add.__init__(
            self, left=kumaraswamy(alpha, beta)*(upper-lower), right=lower)



"""Normal (Gaussian) probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class normal(Dist):
    """Standard normal distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return (2*numpy.pi)**(-.5)*numpy.e**(-x**2/2.)

    def _cdf(self, x):
        return special.ndtr(x)

    def _ppf(self, x):
        return special.ndtri(x)

    def _mom(self, k):
        return .5*special.factorial2(k-1)*(1+(-1)**k)

    def _ttr(self, n):
        return 0., 1.*n

    def _lower(self):
        return -7.5

    def _upper(self):
        return 7.5


class Normal(Add):
    R"""
    Normal (Gaussian) distribution

    Args:
        mu (float, Dist):
            Mean of the distribution.
        sigma (float, Dist):
            Standard deviation.  sigma > 0

    Examples:
        >>> distribution = chaospy.Normal(2, 2)
        >>> distribution
        Normal(mu=2, sigma=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.3168, 1.4933, 2.5067, 3.6832])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.14  , 0.1932, 0.1932, 0.14  ])
        >>> distribution.sample(4).round(4)
        array([ 2.7901, -0.4006,  5.2952,  1.9107])
        >>> distribution.mom(1).round(4)
        2.0
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[ 2.,  2.,  2.],
               [ 4.,  8., 12.]])
    """

    def __init__(self, mu=0, sigma=1):
        self._repr = {"mu": mu, "sigma": sigma}
        Add.__init__(self, left=normal()*sigma, right=mu)



"""Chi distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class chi(Dist):
    """Chi distribution."""

    def __init__(self, df=1):
        Dist.__init__(self, df=df)

    def _pdf(self, x, df):
        return x**(df-1.)*numpy.exp(-x*x*0.5)/(2.0)**(df*0.5-1)\
                /special.gamma(df*0.5)

    def _cdf(self, x, df):
        return special.gammainc(df*0.5,0.5*x*x)

    def _ppf(self, q, df):
        return numpy.sqrt(2*special.gammaincinv(df*0.5, q))

    def _lower(self, df):
        return 0.

    def _mom(self, k, df):
        return 2**(.5*k)*special.gamma(.5*(df+k))\
                /special.gamma(.5*df)


class Chi(Add):
    """
    Chi distribution.

    Args:
        df (float, Dist):
            Degrees of freedom
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Chi(2, 4, 1)
        >>> distribution
        Chi(df=2, scale=4, shift=1)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([ 1.    ,  4.0341,  5.7096,  7.6604, 28.1446])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.    , 0.1422, 0.1472, 0.1041, 0.    ])
        >>> distribution.sample(4).round(4)
        array([ 6.8244,  2.9773, 10.8003,  5.5892])
        >>> distribution.mom(1).round(4)
        6.0133
    """

    def __init__(self, df=1, scale=1, shift=0):
        self._repr = {"df": df, "scale": scale, "shift": shift}
        Add.__init__(self, left=chi(df)*scale, right=shift)


class Maxwell(Add):
    """
    Maxwell-Boltzmann distribution
    Chi distribution with 3 degrees of freedom

    Args:
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Maxwell(2, 3)
        >>> distribution
        Maxwell(scale=2, shift=3)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([ 3.    ,  5.2023,  6.0763,  7.0538, 17.0772])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.    , 0.2638, 0.2892, 0.2101, 0.    ])
        >>> distribution.sample(4).round(4)
        array([6.6381, 4.6119, 8.5955, 6.015 ])
        >>> distribution.mom(1).round(4)
        6.1915
    """

    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=chi(3)*scale, right=shift)


class Rayleigh(Add):
    """
    Rayleigh distribution

    Args:
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Rayleigh(2, 3)
        >>> distribution
        Rayleigh(scale=2, shift=3)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([ 3.    ,  4.5171,  5.3548,  6.3302, 16.5723])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.    , 0.2844, 0.2944, 0.2081, 0.    ])
        >>> distribution.sample(4).round(4)
        array([5.9122, 3.9886, 7.9001, 5.2946])
        >>> distribution.mom(1).round(4)
        5.5066
    """
    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=chi(2)*scale, right=shift)
